#include "astropulse.h"
#include "ap_graphics.h"
#if USE_FFTW
	#include "fftw3.h"
#endif
#include "sbtf.h"
#include "ap_debug.h"
#include <stdio.h>
#include "mtrand.h"
#include <complex>
#if USE_ACML
	#include "acml.h"
#endif
//using std::complex;
//#include "seti_dr2filter.h"

namespace Astropulse {

void Science::print_DC_strengths() {

  const int NUM_STRENGTHS = 8;
  float strengths[NUM_STRENGTHS];
  float totals[NUM_STRENGTHS];
  float averages[NUM_STRENGTHS];
  float maxima[NUM_STRENGTHS];
  // long fft_len = client.ap_shmem->ap_gdata.state.fft_len;
  int fft_len2 = 1024;
  // int fft_len2 = 32768;
  long datasize = 8 * 1024 * 1024;
  FILE* outfile = fopen("temp.txt", "w");
  int num_ffts = 0;

  for(int i = 0; i < NUM_STRENGTHS; i++) {
    totals[i] = 0;
    maxima[i] = 0;
  }

  // Note rawdata is in BYTES, each byte is 4 samples, fft_len is in samples, so # bytes = fft_len2 / 4
  // data_chunk_now is also in bytes, but it INCREMENTS by fft_len2 /2 / 4 to produce overlap.
  // That is, it only increments by HALF of a FFT length each iteration.
  // To find # of samples from data_chunk_now, MULTIPLY BY 4!
  for(long data_chunk_now = 0; data_chunk_now < datasize - fft_len2/2/4; data_chunk_now += fft_len2/2/4) {
    num_ffts++;
#if (__SSE__ && SPLIT_COMPLEX) //uje: && !_GNU_SOURCE
    if (splitter_bits_to_float(&client.rawdata[data_chunk_now], (float *)data, (float *)i_data, fft_len2))

/*#elif _GNU_SOURCE && (__SSE__ && SPLIT_COMPLEX) //uje: for later trying to make SPLIT_COMPLEX work
    if (splitter_bits_to_float(&client.rawdata[data_chunk_now], (float *)data, (float *)i_data, fft_len2))*/

#else
    if (splitter_bits_to_float(&client.rawdata[data_chunk_now],(float *)data, fft_len2))
#endif
	{
      exit(SBTF_ERR);
    }
#if USE_FFTW
    fftwf_execute(fp2); // Note that this fft knows about fft_len2, and computes in chunks of fft_len2
			   // The maximum possible result is 1,024, average should be such that power = 1024
#elif USE_ACML
	cfft1d(-1,acmlfftlen2,(complex*)data,fp2,&info);
#endif
    for(int i = 0; i < NUM_STRENGTHS; i++) {
      strengths[i] = 0;
      long amt_to_subtract = (i==0)?0:pow_int(2,i-1);
      for(long j = 0; j < pow_int(2,i); j++) {
        float real = ((ap_complex *) data)[(j - amt_to_subtract + fft_len2) % fft_len2][0];
        float imag = ((ap_complex *) data)[(j - amt_to_subtract + fft_len2) % fft_len2][1];
        strengths[i] += real * real + imag * imag;
      }
      totals[i] += strengths[i];
      maxima[i] = (strengths[i] > maxima[i])?strengths[i]:maxima[i];
      // fprintf(outfile, "strengths[%d] = %f\n", i, strengths[i]);
      // fprintf(outfile, "totals[%d] = %f\n", i, totals[i]);
      // fprintf(outfile, "maxima[%d] = %f\n", i, maxima[i]);
      // fprintf(outfile, "strengths[%d] / %d = %f\n", i, pow_int(2, i), strengths[i] / pow(2, i));
      fflush(outfile);
    }
  }

  fprintf(outfile, "The averages below represent power per DC bin.\n");
  for(int i = 0; i < NUM_STRENGTHS; i++) {
    averages[i] = totals[i] / (float) num_ffts / (float) pow_int(2,i) / (float) fft_len2 / 2.0;
    fprintf(outfile, "averages[%d] = %f\n", i, averages[i]);
  }
  for(int i = 0; i < NUM_STRENGTHS; i++) {
    fprintf(outfile, "normalized maxima[%d] = %f\n", i, maxima[i] / (float) pow_int(2,i) / (float) fft_len2 / 2.0);
  }
  fflush(outfile);
}

int Science::get_indices_to_randomize(std::vector<long>& indices, int scale) {
  FILE *outfile;
  int init_indices = 100;
  int num_indices = 0;
  int size = init_indices;
  indices.resize(init_indices);
  if(scale > 7) {
      fprintf(stderr, "In Science::get_indices_to_randomize: error: scale > 7\n");
      printf("In Science::get_indices_to_randomize: error: scale > 7\n");
      exit(-1);
  }
  // power_threshold[scale]
  float power_threshold[] = { 12.1, 6.9, 5.3, 3.0, 2.46, 1.96, 1.62, 1.37};
  int fft_len2 = 1024;
  long datasize = 8 * 1024 * 1024;
  float strength;
  outfile = fopen("indices.txt", "w");

  for(long data_chunk_now = 0; data_chunk_now < datasize - fft_len2/2/4; data_chunk_now += fft_len2/2/4) {
#if (__SSE__ && SPLIT_COMPLEX) //uje: && !_GNU_SOURCE
    if (splitter_bits_to_float(&client.rawdata[data_chunk_now], (float *)data, (float *) i_data,  fft_len2))

/*#elif _GNU_SOURCE && (__SSE__ && SPLIT_COMPLEX) //uje: for later trying to make SPLIT_COMPLEX work
    if (splitter_bits_to_float(&client.rawdata[data_chunk_now], (float *)data, (float *) i_data,  fft_len2))*/

#else
    if (splitter_bits_to_float(&client.rawdata[data_chunk_now], (float *)data, fft_len2))
#endif
	{
      exit(SBTF_ERR);
    }
#if USE_FFTW
    fftwf_execute(fp2); // Note that this fft knows about fft_len2 (if you define it in astropulse.h), and computes in chunks of fft_len2
			   // The maximum possible result is 1,024, average should be such that power = 1024
#endif
#if USE_ACML
	cfft1d(-1,acmlfftlen2,(complex*)data,fp2,&info);
#endif
    /* For debugging: check power in DC bin for this data chunk
    if(data_chunk_now == 1359872) {
        printf("DC bin: %f %f, DC power: %f\n", data[0][0], data[0][1], (data[0][0] * data[0][0] + data[0][1] * data[0][1]) / fft_len2 / 2.0);
    }*/

    int expscale = pow_int(2, scale);
    strength = 0;
    for(int j = - expscale / 2; j < (expscale + 1) / 2; j++) {
      int j2 = (j + fft_len2) % fft_len2;
#if !(_GNU_SOURCE || __FreeBSD__ || __APPLE_CC__)
      float real = data[j2][0];
      float imag = data[j2][1];
#elif _GNU_SOURCE || __FreeBSD__ || __APPLE_CC__
      float real = ((ap_complex*)data)[j2][0];
      float imag = ((ap_complex*)data)[j2][1];
#endif
      strength += (real * real + imag * imag) / fft_len2 / 2.0 / expscale;
      // strength += (real * real + imag * imag) / fft_len2 / expscale;
    }

    /* For debugging: check that total power / 2.0 = fft_len2
    if(data_chunk_now < 10000) {
    float total = 0;
    for(int i = 0; i < fft_len2; i++) {
        total += (data[i][0] * data[i][0] + data[i][1] * data[i][1]) / fft_len2 / 2.0;
    }
    printf("total = %f\n", total);
    }
    */

    if(strength > power_threshold[scale]) {
      fprintf(outfile, "At sample %ld, data_chunk_now %ld, strength %f > %f\n", data_chunk_now * 4, data_chunk_now, strength, power_threshold[scale]);
      /* Given 3 index finds, if the blanking range of the first overlaps the range
       * of the third, the one in the middle is redundant so overwrite it rather than
       * extending the list.
       */
      if(num_indices > 1 && (indices[num_indices - 2] + 100000) >= (data_chunk_now - 100000))
        num_indices--;
      indices[num_indices++] = data_chunk_now;
      if(num_indices >= size) {
        size = num_indices + init_indices;
        indices.resize(size);
      }
    }
  }
  if (num_indices) {
    fprintf(outfile, "\nRetained data_chunk indices:\n");
    for (int k = 0; k < num_indices; k++) fprintf(outfile, "%ld\n", indices[k]);
  }
  fflush(outfile);
  fclose(outfile);
  return num_indices;
}

// Read in as many data chunks as needed from client.rawdata.
// Warning: this function messes with the Science data array.
void Science::generate_envelope(const std::vector<long>& indices, int num_indices) {
  // initialize the envelope array
  for(int i=0; i<fftlen_envelope; i++)
     envelope[i] = 0;

  // GOAL: find envelope (as it should be after one-bitting)

  // For counting to see if we've performed enough ffts
  // for a meaningful power envelope.
  int num_ffts_performed = 0;

  // Check whether envelope size is (wrongly) larger than fft len
  // This should never happen, but we check just in case.
  if(fftlen_envelope > client.ap_shmem->ap_gdata.state.fft_len) {
    printf("Error in ap_remove_radar.cpp : Science::generate_envelope : fftlen_envelope is > client fft_len and should not be.\n");
    fprintf(stderr, "Error in ap_remove_radar.cpp : Science::generate_envelope : fftlen_envelope is > client fft_len and should not be.\n");
  }

  // * cycle through the #s in the data, adding each of them to
  //   envelope_part
  //   env_index tells which envelope_part we are on
  bool good_data_chunk;
  int data_chunk_size = client.ap_shmem->ap_gdata.state.fft_len / 4;
  for(int data_chunk=0; (unsigned)data_chunk <= client.rawdata.size() - data_chunk_size; data_chunk+= data_chunk_size) {

    // Check whether this is a bad data chunk (within 100000 bytes of RFI.)
    good_data_chunk = true;
    for(int n=0; n<num_indices; n++) {
      if( indices[n] < data_chunk + 100000 && indices[n] > data_chunk - 100000) {
        good_data_chunk = false;
	break;
      }
    }
    if(!good_data_chunk) continue;

    // Extract our data chunk into the data array.
    splitter_bits_to_float(&client.rawdata[data_chunk], (float*)data, client.ap_shmem->ap_gdata.state.fft_len);
    for(int env_index=0; env_index<client.ap_shmem->ap_gdata.state.fft_len / fftlen_envelope; env_index++) {
		for(int i = 0; i < fftlen_envelope; i++) {
		  int d = env_index * fftlen_envelope + i;
#if !(_GNU_SOURCE || __FreeBSD__ || __APPLE_CC__)
		  envelope_part[i][0] = data[d][0];
		  envelope_part[i][1] = data[d][1];
#elif _GNU_SOURCE || __FreeBSD__ || __APPLE_CC__
		  envelope_part[i][0] = ((ap_complex*)data)[d][0];
		  envelope_part[i][1] = ((ap_complex*)data)[d][1];
#endif
		}
#if USE_FFTW
      fftwf_execute(fp_envelope);
#endif
#if USE_ACML
	  cfft1d(-1,acmlfftlenenv,(complex*)envelope_part,fp_envelope,&info);
#endif
      for(int i = 0; i < fftlen_envelope; i++) {
        envelope[i] += envelope_part[i][0] * envelope_part[i][0]
                     + envelope_part[i][1] * envelope_part[i][1];
      }
      num_ffts_performed++;
    }
  }

  // * if you don't get very many envelope_part's, print an error msg and exit.
  //   (this means there was too much RFI.)
  int min_ffts = 100;
  if(num_ffts_performed < min_ffts) {
    printf("Error in ap_remove_radar.cpp : generate_envelope: num_ffts_performed < %d.  Blanking too much RFI?\n", min_ffts);
    fprintf(stderr, "Error in ap_remove_radar.cpp: generate_envelope: num_ffts_performed < %d.  Blanking too much RFI?\n", min_ffts);
    // Quit early
    fflush(stdout);
    client.outfile.write_main(client.signal_vector);
    client.outfile.finish_results();
    client.finish();
    main_timer.stop();
    boinc_finish(0);
    exit(0);
  }

  // * when you're done, normalize the power array:
  //    * add up the powers, and divide by fftlen_envelope
  //      to get the average power
  //    * divide the whole array by the average power
  //    * array is avp(nu), the average power at frequency nu, normalized
  //      so that <avp(nu)>_nu = 1.
  double total_power = 0;
  for(int i = 0; i < fftlen_envelope; i++)
    total_power += envelope[i];
  double avg_power = total_power / fftlen_envelope;
  for(int i = 0; i < fftlen_envelope; i++)
    envelope[i] /= avg_power;
}

void Science::generate_pre_envelope() {

  double known_alpha = 0.547539;

  double total_power=0; //uje: preventing uninitialized usage
  for(int i = 0; i < fftlen_envelope; i++) {
    pre_envelope[i] = (1 + known_alpha) * envelope[i] - known_alpha;
    if(pre_envelope[i] < 0) pre_envelope[i] = 0;
    total_power += pre_envelope[i];
  }
  if(total_power <= 0) {
     printf("Error in ap_remove_radar.cpp: generate_pre_envelope: total <= 0.\n");
     fprintf(stderr, "Error in ap_remove_radar.cpp: generate_pre_envelope: total <= 0.\n");
     exit(1);
  }
  for(int i = 0; i < fftlen_envelope; i++) // set average power per bin to 1.
    pre_envelope[i] *= (fftlen_envelope / total_power);
}

void Science::randomize_indices(const std::vector<long>& indices, int num_indices, int data_chunk_now, int fft_len, state_t& state) {

// FILE* randsampfile = fopen("randsamp.txt", "a");
  int dcn_up  = data_chunk_now + 100000;
  int dcn_dwn = data_chunk_now - 100000;

  /* For debugging: print values of bits before they are randomized.
  if(data_chunk_now == 1359872 - 4096) {
      printf("data_chunk_now: 1359872 - 4096\n");
      for(int k = 16384; k < 16384 + 1024; k++) {
          printf("%f %f | ", data[k][0], data[k][1]);
          if(k % 16 == 0) printf("\n");
      }
  }
  */

  // printf("data_chunk_now = %d\n", data_chunk_now); // for debugging

  // The count of data chunks blanked divided by all data chunks is blanking ratio.
  state.lastDataChunkBlankedNew++ ; //increment all count

  for(int current = 0; current < num_indices; current++) {
      // If we're within 100000 bytes of a DC pulse, randomize this data chunk.
	  if(indices[current] < dcn_up && indices[current] > dcn_dwn) {
      int debug = 0;
      state.numSamplesBlankedNew++ ; //increment blanked count
      randomize_samples(fft_len, state.mtr_new, debug);
      // fprintf(randsampfile, "Num chunks blanked: %d, Total chunks: %d, Percent blanked: %f\n",
      // state.numSamplesBlankedNew, state.lastDataChunkBlankedNew,
      // 100.0f * (float)state.numSamplesBlankedNew / (float)state.lastDataChunkBlankedNew);
      break;
    }
  }
  // fclose(randsampfile);
}
#ifdef USE_UBUNTU904
static const double sqrbs[] =
{
    3.162277660168379331998893544433, 1.778279410038922801225421195193, 1.333521432163324025675931715295,
    1.154781984689458179666482887296, 1.074607828321317497215941531964, 1.036632928437697997291651724925,
    1.018151721718181841474226888579, 1.009035044841447437759254423906, 1.004507364254462515664794694341,
    1.002251148292912915465673638867, 1.001124941399879875885426434366, 1.00056231260220863661851136781,
    1.00028111678778013239925736577,  1.000140548516947258162771187859, 1.000070271789411435538813638677,
    1.000035135277461856608582335862, 1.000017567484422673833847265274, 1.000008783703634612146574315569,
    1.000004391842173167236282001464, 1.000002195918675554203317137505, 1.000001097958735020409754759572,
    1.000000548979216821114626617594, 1.000000274489570738295091262044, 1.000000137244775951083282699498,
    1.000000068622385621025737189369, 1.000000034311192221883912751151, 1.000000017155595963784719939262,
    1.000000008577797945103051176124, 1.000000004288898963354198429132, 1.000000002144449479377767429764,
    1.000000001072224739114050769298, 1.00000000053611236941331714833,  1.000000000268056184670731515095,
    1.000000000134028092326383992781, 1.000000000067014046160946555199, 1.000000000033507023079911917302,
    1.000000000016753511539815618576, 1.00000000000837675576987272427,  1.00000000000418837788492759088,
    1.000000000002094188942461602626, 1.00000000000104709447123025311,  1.000000000000523547235614989504,
    1.000000000000261773617807460489, 1.000000000000130886808903721679, 1.000000000000065443404451858698,
    1.000000000000032721702225928814, 1.000000000000016360851112964273, 1.000000000000008180425556482103,
    1.000000000000004090212778241043, 1.000000000000002045106389120519, 1.000000000000001022553194560259,
    1.000000000000000511276597280129, 1.000000000000000255638298640065, 1.000000000000000127819149320032,
};
static const double lnrts[] =
{
    0.5,                      0.25,                      0.125,
    0.0625,                   0.03125,                   0.015625,
    0.0078125,                0.00390625,                0.001953125,
    0.0009765625,             0.00048828125,             0.000244140625,
    0.0001220703125,          0.00006103515625,          0.000030517578125,
    0.0000152587890625,       0.00000762939453125,       0.000003814697265625,
    0.0000019073486328125,    0.00000095367431640625,    0.000000476837158203125,
    0.0000002384185791015625, 0.00000011920928955078125, 0.000000059604644775390625,
    0.0000000298023223876953125,           0.00000001490116119384765625,
    0.000000007450580596923828125,         0.0000000037252902984619140625,
    0.0000000018626451492309570312,        0.00000000093132257461547851562,
    0.00000000046566128730773925781,       0.00000000023283064365386962891,
    0.00000000011641532182693481445,       0.000000000058207660913467407227,
    0.000000000029103830456733703613,      0.000000000014551915228366851807,
    0.0000000000072759576141834259033,     0.0000000000036379788070917129517,
    0.0000000000018189894035458564758,     0.00000000000090949470177292823792,
    0.00000000000045474735088646411896,    0.00000000000022737367544323205948,
    0.00000000000011368683772161602974,    0.00000000000005684341886080801487,
    0.000000000000028421709430404007435,   0.000000000000014210854715202003717,
    0.0000000000000071054273576010018587,  0.0000000000000035527136788005009294,
    0.0000000000000017763568394002504647,  0.00000000000000088817841970012523234,
    0.00000000000000044408920985006261617, 0.00000000000000022204460492503130808,
    0.00000000000000011102230246251565404, 0.000000000000000055511151231257827021,
};

inline double lgnbriggs(double arg){
    //double lnrts[53], sqrbs[53];
    double res=0.0, imd=1.0, ret=1.0; //lgr=0.5;
    int ct=0, stl=0;
    /* uje: precalculated the two tables above this way :
    for(ct;ct<=53;ct++){
        lnrts[ct]=lgr;
        lgr/=2;
    }
    sqrbs[0]=sqrt(10);
    for(ct=1;ct<=53;ct++){
        sqrbs[ct]=sqrt(sqrbs[ct-1]);
        //printf("sqrbs[%i] = %-.30g , lnrts[%i] = %-.30g  \n", ct, sqrbs[ct], ct-1, lnrts[ct-1]);
    }*/
    if( arg<1){
        do{
            stl--;
            arg*=10;
        }while(arg<1);
    }
    //printf("arg = %-.30g , stl = %i , res = %-.30g \n", arg, stl, res);
    for(ct=0;ct<=15;ct++){ //uje: 15 steps is enough in standalone for "strongly similar" results
        imd*=sqrbs[ct];
        (imd<arg) ? res=res+lnrts[ct] : imd=ret;
        ret=imd;
        //printf("ct = %i, imd = %-.30g , arg = %-.30g , res = %-.30g \n", ct, imd, arg, res);
    }
    res+=((float)stl);
    //printf("res = %-.30g \n", res);
    res*=2.30258509299404574224978023654; // uje: constant calculating log_e(x)
    return res;
}

double gasdev(MtRand &mtr) {
  gasdev_ct++;
  static int iset=0;
  static double gset;
  double fac, rsq, v1, v2;
  if(iset == 0) {
    do {
      v1=2.0*mtr.drand32()-1.0;
      v2=2.0*mtr.drand32()-1.0;
      rsq=v1*v1+v2*v2;
    } while(rsq >= 1.0 || rsq==0.0);
    gasdev_sqrtlog_ct++;
    gasdev_minlog = (rsq<gasdev_minlog)? rsq : gasdev_minlog; // min>0
    gasdev_maxlog = (rsq>gasdev_maxlog)? rsq : gasdev_maxlog; // max<=1
    if(rsq>0.0) {
        fac=(1/rsq)*sqrt(-2.0*lgnbriggs(rsq)*rsq);
    } else {
        fprintf(stderr,"gasdev, log() avoiding SINGularity error.");
        fac=(1/(rsq+0.00000000000001))*sqrt(-2.0*lgnbriggs(rsq+0.00000000000001)*rsq);
    }
    gset=v1*fac;
    iset=1;
    return v2*fac;
  } else {
    iset=0;
    return gset;
  }
}
#else
double gasdev(MtRand &mtr) {
    //gasdev_ct++;
    static int iset=0;
    static double gset;
    double fac, rsq, v1, v2;
    if(iset == 0) {
        do {
            //	  v1=2.0*mtr.drand32()-1.0;
			//	  v2=2.0*mtr.drand32()-1.0;
			//JWS: retrieve as 32 bits of unsigned long. Toggling the 2^31 bit and casting to
			// signed integer gives -2^31 to +2^31-1 range, replacing the -1.0 offsetting
			// operation. The xor is faster, so is converting from signed integer to double.
			v1=((int)(mtr.random32()^0x80000000))*(1.0/2147483648.0);
			v2=((int)(mtr.random32()^0x80000000))*(1.0/2147483648.0);
            rsq=v1*v1+v2*v2;
        } while(rsq >= 1.0 || rsq==0.0);
        //gasdev_sqrtlog_ct++;
        //gasdev_minlog = (rsq<gasdev_minlog)? rsq : gasdev_minlog; // min>0
        //gasdev_maxlog = (rsq>gasdev_maxlog)? rsq : gasdev_maxlog; // max<=1
        if(rsq>0.0) {
            fac=(1/rsq)*sqrt(-2.0*log(rsq)*rsq);
        } else {
            fprintf(stderr,"gasdev, log() avoiding SINGularity error.");
            fac=(1/(rsq+0.00000000000001))*sqrt(-2.0*log(rsq+0.00000000000001)*rsq);
        }
        gset=v1*fac;
        iset=1;
        return v2*fac;
    } else {
        iset=0;
        return gset;
    }
}
#endif //USE_UBUNTU904
// One-bits the data (called "noise") which has length fftlen_noise
// That is, values greater than 0 are set to 1, and less than 0 are
// set to -1.
void one_bit(fftwf_complex * noise, int fftlen_noise) {
  int sum_0_0 = 0;
  int sum_0_1 = 0;
  for(int k=0; k<fftlen_noise; k++) {
    noise[k][0] = (noise[k][0] < 0) ? -1 : 1;
    noise[k][1] = (noise[k][1] < 0) ? -1 : 1;
    sum_0_0 += noise[k][0];
    sum_0_1 += noise[k][1];
  }
  // printf("noise[0][0] = %f, noise[0][1] = %f\n", noise[0][0], noise[0][1]);
  // printf("sum_0_0 = %d, sum_0_1 = %d\n", sum_0_0, sum_0_1);
}

// This generates unnormalized time domain noise with frequency spectrum given by "envelope".
// (Because it will be one-bitted immediately afterwards.)
#if USE_FFTW
void generate_time_noise_with_envelope(fftwf_complex * noise, double * envelope, int fftlen_noise, int fftlen_envelope, fftwf_plan fp_noise_inv, MtRand &mtr) {
  // double rnd1;
  // double rnd2;
  // double HALF_RAND_MAX = RAND_MAX/2.0;
  double multiplier;

  //  * generate fftlen_patch random cplx bit pairs (as doubles)
  /*
  for(int k=0; k<fftlen_noise; k++) {
    rnd1=static_cast<double>(rand());
    rnd1=rnd1 <= HALF_RAND_MAX ? -1 : 1;
    rnd2=static_cast<double>(rand());
    rnd2=rnd2 <= HALF_RAND_MAX ? -1 : 1;
    noise[k][0]=rnd1;
    noise[k][1]=rnd2;
  }*/
  //  * fourier transform them to get a Gaussian distribution
  //      average power per frequency bin is some constant P, and
  //      expected power in each bin is <P>.
  // fftwf_execute(fp_noise);
  // avg power per bin = 2 * fftlen_patch.
  //  * TODO: divide by sqrt(P), so avg power in each bin is 1.
  /* double sqrt_avg_power = sqrt((double) 2 * fftlen_noise);
  for(int k = 0; k<fftlen_noise; k++) {
    noise[k][0] /= sqrt_avg_power;
    noise[k][1] /= sqrt_avg_power;
  } */
  for(int k=0; k<fftlen_noise; k++) {
    noise[k][0] = gasdev(mtr);
    noise[k][1] = gasdev(mtr);
  }
  //  * multiply patch by sqrt(avp(nu)).
  //      expected power in bin nu is now avp(nu)
  for(int k = 0; k<fftlen_noise; k++) {
    // 0.5 is added so we get 0.5, ... fftlen_noise - 0.5, which is then rounded down.
    int index = (int) ( ((float)k + 0.5) / (float)fftlen_noise * (float)fftlen_envelope);
    multiplier = sqrt(envelope[ index ]);

    noise[k][0] *= multiplier;
    noise[k][1] *= multiplier;
    // if(k == 0)
    //  printf("noise[0][0] = %f, noise[0][1] = %f, noise power = %f, multiplier = %f, index = %d\n", noise[0][0], noise[0][1], noise[0][1] * noise[0][1] + noise[0][0] * noise[0][0], multiplier, index);
  }
  //  * inverse FFT
  fftwf_execute(fp_noise_inv);
}
#endif
#if USE_ACML
void generate_time_noise_with_envelope(ap_complex * noise, double * envelope, int fftlen_noise, int fftlen_envelope, complex * fp_noise_inv, MtRand &mtr) {

  // double rnd1;
  // double rnd2;
  // double HALF_RAND_MAX = RAND_MAX/2.0;
  double multiplier;
  int info;

  //  * generate fftlen_patch random cplx bit pairs (as doubles)
  /*
  for(int k=0; k<fftlen_noise; k++) {
    rnd1=static_cast<double>(rand());
    rnd1=rnd1 <= HALF_RAND_MAX ? -1 : 1;
    rnd2=static_cast<double>(rand());
    rnd2=rnd2 <= HALF_RAND_MAX ? -1 : 1;
    noise[k][0]=rnd1;
    noise[k][1]=rnd2;
  }*/
  //  * fourier transform them to get a Gaussian distribution
  //      average power per frequency bin is some constant P, and
  //      expected power in each bin is <P>.
  // fftwf_execute(fp_noise);
  // avg power per bin = 2 * fftlen_patch.
  //  * TODO: divide by sqrt(P), so avg power in each bin is 1.
  /* double sqrt_avg_power = sqrt((double) 2 * fftlen_noise);
  for(int k = 0; k<fftlen_noise; k++) {
    noise[k][0] /= sqrt_avg_power;
    noise[k][1] /= sqrt_avg_power;
  } */
  for(int k=0; k<fftlen_noise; k++) {
    noise[k][0] = gasdev(mtr);
    noise[k][1] = gasdev(mtr);
  }
  //  * multiply patch by sqrt(avp(nu)).
  //      expected power in bin nu is now avp(nu)
  for(int k = 0; k<fftlen_noise; k++) {
    // 0.5 is added so we get 0.5, ... fftlen_noise - 0.5, which is then rounded down.
    int index = (int) ( ((float)k + 0.5) / (float)fftlen_noise * (float)fftlen_envelope);
    multiplier = sqrt(envelope[ index ]);

    noise[k][0] *= multiplier;
    noise[k][1] *= multiplier;
    // if(k == 0)
    //  printf("noise[0][0] = %f, noise[0][1] = %f, noise power = %f, multiplier = %f, index = %d\n", noise[0][0], noise[0][1], noise[0][1] * noise[0][1] + noise[0][0] * noise[0][0], multiplier, index);
  }
  //  * inverse FFT
  //fftwf_execute(fp_noise_inv);
  cfft1d(1,fftlen_noise,(complex *)noise,fp_noise_inv,&info);
}
#endif

// Use debug flag if you want randomize_samples to print something out
// on specific samples, for instance.
void Science::randomize_samples(int fft_len, MtRand& mtr, int debug) {
  // FILE *randsampfile;
  // randsampfile = fopen("randsamp.txt", "a");
  // fprintf(randsampfile, "in randomize_samples: in: %s\n", mtr.toString());
  int patch_index = 0;

  for(int index = 0; index < fft_len; index++) {
    if(patch_index == 0) {
      generate_time_noise_with_envelope(patch, pre_envelope, fftlen_patch, fftlen_envelope, fp_patch_inv, mtr);
      one_bit(patch, fftlen_patch);
    }
// * use bits from this distribution to fill the holes
    while(index < client.ap_shmem->ap_gdata.state.fft_len) {
#if !(_GNU_SOURCE || __FreeBSD__ || __APPLE_CC__)
      data[index][0] = patch[patch_index][0];
      data[index][1] = patch[patch_index][1];
#elif _GNU_SOURCE || __FreeBSD__ || __APPLE_CC__
      ((ap_complex*)data)[index][0] = ((ap_complex*)patch)[patch_index][0];
      ((ap_complex*)data)[index][1] = ((ap_complex*)patch)[patch_index][1];
#endif
    // data[index][0] = mtr.drand32() > 0.5 ? 1.0 : - 1.0;
    // data[index][1] = mtr.drand32() > 0.5 ? 1.0 : - 1.0;
      patch_index++;
      if(patch_index >= fftlen_patch) {
// * when you don't have any patch bits left,
// start a new patch, going back through the for loop
        patch_index = 0;
        break;
      }
      index++;
    }
    patch_index = 0;
  } // end "for(int index = 0; index < fft_len; index++)"
  // fprintf(randsampfile, "in randomize_samples: out: %s\n", mtr.toString());
  // fflush(randsampfile);
  // fclose(randsampfile);
}

} // end namespace Astropulse
